/////////////////////////////////////////////////////////////////////
//
// RC Library, (c) 2011-2014, Ryan A. Colyer
// Distributed under the Boost Software License, v1.0. (LICENSE.txt)
//
/// \file RCBits.h
/// Provides short convenience classes and functions.
/////////////////////////////////////////////////////////////////////

#ifndef RC_RCBITS_H
#define RC_RCBITS_H

#include "RCconfig.h"
#ifdef CPP11
#include "Tuple.h"
#endif

namespace RC {

/// @cond UNDOC
template<class T>
class BetwCompare {
  protected:
  const T& x;
  bool ans;
  public:
  inline explicit BetwCompare(const T& x) : x(x), ans(true) { }
#define RC_BETW_HELP(REL) \
  template<class T2> \
  inline BetwCompare operator REL (const T2& other) { \
    ans &= x REL other; \
    return *this; \
  }
  RC_BETW_HELP(<)
  RC_BETW_HELP(>)
  RC_BETW_HELP(<=)
  RC_BETW_HELP(>=)
  RC_BETW_HELP(==)
  RC_BETW_HELP(!=)
  inline operator bool () const { return ans; }
};

#define RC_BETW_HELP2(REL1, REL2) \
template<class T, class T2> \
inline BetwCompare<T> operator REL1 (const T2& other, BetwCompare<T> betw) { \
  return betw REL2 other; \
}

RC_BETW_HELP2(<, >)
RC_BETW_HELP2(>, <)
RC_BETW_HELP2(<=, >=)
RC_BETW_HELP2(>=, <=)
RC_BETW_HELP2(==, ==)
RC_BETW_HELP2(!=, !=)

/// @endcond

/// Returns a comparator that can be used for range comparisons.
/** Usage exmple:  if (3.14 < Betw(5) <= 7) { cout << "Yes\n"; }
 *  Works for any comparable types, as long as one of the left-most two in
 *  the chain is wrapped in Betw.
 */
template<class T>
BetwCompare<T> Betw(const T& x) { return BetwCompare<T>(x); }

#ifdef CPP11

/// @cond UNDOC
template<class... Args>
class OneOfCompare {
  protected:
  Tuple<Args...> compare;
  public:
  OneOfCompare(Tuple<Args...>&& tup) : compare(tup) { }
  template<class T, class... Args2>
  friend bool operator==(const T& item, const OneOfCompare<Args2...>& list);
  template<class T, class... Args2>
  friend bool operator!=(const T& item, const OneOfCompare<Args2...>& list);
};

template<class T, class... Args>
bool operator== (const T& item, const OneOfCompare<Args...>& list) {
  Data1D<T> arr = list.compare.template AsData<T>();
  for (T& elem : arr) {
    if (item == elem) {
      return true;
    }
  }
  return false;
}
template<class T, class... Args>
bool operator!= (const T& item, const OneOfCompare<Args...>& list) {
  return ! (item == list);
}
/// @endcond

/// Returns a comparator that can be used to check if something is in a list.
/** Each element given as a parameter to OneOf is constructed with the type
 *  of the object left of the == or != operator.  Usage exmple:
 *  RStr str = "foo";  if (str != OneOf("bar", -3.54, "blah")) {
 *  cout << "Not Found\n"; }  int x = 8; if (5 == OneOf(3.14, 5, x)) {
 *  cout << "Found\n"; }
 */
template<class... Args>
OneOfCompare<Args...> OneOf(Args... args) {
  return OneOfCompare<Args...>(Tuple<Args...>(args...));
}
#endif

}

#endif // RC_RCBITS_H

